import pygame
import numpy as np

from gymnasium.spaces import Discrete, Box

from sai_pygame.utils.env import ArenaXGameBase
from sai_pygame.utils.colors import WHITE, BLACK

from .entities.paddle import Paddle
from .entities.ball import Ball

# Game Constants
PADDLE_START_X = 20
PADDLE_START_Y = 200
PADDLE_SPEED = 5
MAX_VELOCITY = 7

# Config
SCREEN_WIDTH = 704
SCREEN_HEIGHT = 480
FRAMERATE = 60

# Action Mapping
ACTION_MAPPING = {
    "up": (pygame.K_w, 1),
    "down": (pygame.K_s, 2),
}


class PongEnv(ArenaXGameBase):
    metadata = {
        "render_modes": ["human", "rgb_array"],
        "render_fps": FRAMERATE,
        "width": SCREEN_WIDTH,
        "height": SCREEN_HEIGHT,
        "engine": "pygame",
        "reward_functions": ["classic"],
    }

    def __init__(
        self,
        width=SCREEN_WIDTH,
        height=SCREEN_HEIGHT,
        framerate=FRAMERATE,
        render_mode="rgb_array",
        reward_function="classic",
        max_velocity=MAX_VELOCITY,
        seed=None,
        **kwargs,
    ):
        super().__init__(
            width,
            height,
            framerate=framerate,
            render_mode=render_mode,
            game_name="Pong - ArenaX Labs",
            action_mapping=ACTION_MAPPING,
            seed=seed,
            **kwargs,
        )
        # initialize custom variables
        self.TIME_LIMIT = 120
        self.ball_size = self.screen_width / 70
        self.paddle_width = self.screen_width / 70
        self.paddle_height = self.screen_height / 5

        self.max_velocity = max_velocity

        # Level based progression
        self.level = 1
        self.base_opponent_speed = 1
        self.current_opponent_speed = self.base_opponent_speed
        self.speed_increase_per_level = 1

        self.winner = ""

        # initialize paddles and ball
        self.paddleA = Paddle(
            WHITE,
            self.paddle_width,
            self.paddle_height,
            self.screen_height,
            PADDLE_SPEED,
        )
        self.paddleB = Paddle(
            WHITE,
            self.paddle_width,
            self.paddle_height,
            self.screen_height,
            self.current_opponent_speed,
        )
        self.ball = Ball(
            self,
            WHITE,
            self.ball_size,
            self.screen_width,
            self.screen_height,
            self.max_velocity,
        )

        # create a group of sprites for easier updating and rendering
        self.all_sprites_list = pygame.sprite.Group()
        self.all_sprites_list.add(self.paddleA, self.paddleB, self.ball)

        # reset game
        self.reset()

        # Define action and observation space
        self.action_space = Discrete(3)
        self.observation_space = Box(
            low=np.array(
                [-1, -1, -1, -1, -1, -1]
            ),  # Normalized values between -1 and 1
            high=np.array([1, 1, 1, 1, 1, 1]),
            dtype=np.float32,
        )

        # initialize reward function
        self.reward_functions = {
            "classic": self._classic_reward,
        }
        self.reward_function = self.reward_functions[reward_function]
        assert self.reward_function is not None, (
            f"Reward function {reward_function} not found"
        )

        # initialize observation
        self.init_obs = self.get_observation()

    def reset(self, **kwargs):
        """
        Reset the game to its initial state.
        """
        super().reset(**kwargs)

        # reset the game environment back to the initial state
        self.level = 1
        self.current_opponent_speed = self.base_opponent_speed
        self.paddleB.speed = self.current_opponent_speed
        self.winner = ""

        # Position paddles and ball to their start positions
        self._reset_positions()

        observation = self.get_observation()

        info = {
            "timestep": self.frame,
            "screen_height": self.screen_height,
            "screen_width": self.screen_width,
            "paddle_height": self.paddle_height,
            "ball_position": self.ball.rect.center,
            "left_paddle_position": self.paddleA.rect.center,
            "right_paddle_position": self.paddleB.rect.center,
        }

        return observation, info

    def move_opponent(self):
        """
        Move the opponent paddle towards the ball.
        """

        if self.np_random.random() < 0.10:  # 10% chance to make mistake
            return

        ball_center_y = self.ball.rect.y + self.ball.diameter / 2

        aim_position = self.np_random.random()  # Random float between 0.0 and 1.0
        aim_position = max(0.1, min(0.9, aim_position))

        # Calculate the target Y position on the paddle
        target_paddle_y = self.paddleB.rect.y + (aim_position * self.paddleB.height)

        if ball_center_y < target_paddle_y:
            self.paddleB.moveUp()
        else:
            self.paddleB.moveDown()

    def check_collisions(self):
        """
        Check for collisions with walls and paddles, adjust game state accordingly.
        """
        # Wall collision
        if self.ball.rect.x >= self.screen_width - self.ball.diameter:
            self.winner = "A"
        elif self.ball.rect.x <= 0:
            self.done = True
            self.winner = "B"

        if (
            self.ball.rect.y >= self.screen_height - self.ball.diameter
            or self.ball.rect.y <= 0
        ):
            self.ball.velocity[1] = -self.ball.velocity[1]

        # Paddle collision
        if pygame.sprite.collide_mask(self.ball, self.paddleA):
            self.ball.bounce(self.paddleA)

        if pygame.sprite.collide_mask(self.ball, self.paddleB):
            self.ball.bounce(self.paddleB)

    def step(self, action):
        """
        Perform one step of the game, then extract the observation.
        """
        self.update(action)
        super().step()

        observation = self.get_observation()

        done = self.done
        truncated = self.truncated
        reward = self.get_reward()
        if self.winner == "A":
            self.increase_difficulty()

        info = {
            "timestep": self.frame,
            "screen_height": self.screen_height,
            "screen_width": self.screen_width,
            "paddle_height": self.paddle_height,
            "ball_position": self.ball.rect.center,
            "left_paddle_position": self.paddleA.rect.center,
            "right_paddle_position": self.paddleB.rect.center,
        }

        return observation, reward, done, truncated, info

    def update(self, action):
        """
        Move the game state forward one frame, given an action for the paddle(s).
        """
        # game loop goes here

        # Update all sprites
        self.all_sprites_list.update()

        # Move paddles according to the action
        if action == 1:
            self.paddleA.moveUp()
        elif action == 2:
            self.paddleA.moveDown()

        # Ball and wall collision logic
        self.check_collisions()
        if self.TIME_LIMIT - self.time <= 0:
            self.truncated = True

        # Move opponent paddle
        self.move_opponent()

    def render(self):
        """
        Render the game based on the mode.
        """
        # draw everything to the screen
        self.screen.fill(BLACK)
        self.all_sprites_list.draw(self.screen)
        return super().render()

    def get_paddle_position(self, y):
        """Convert paddle y position to normalized coordinates between -1 and 1"""
        # Remove the scaling factor and simply normalize to [-1, 1]
        return (y / self.screen_height - 0.5) * 2

    def _reset_positions(self):
        """
        Reset the positions of the paddles and ball to their initial state.
        """
        self.winner = ""
        self.paddleA.rect.y = PADDLE_START_Y
        self.paddleB.rect.y = PADDLE_START_Y
        self.paddleB.rect.x = self.screen_width - PADDLE_START_X - self.paddle_width
        self.paddleA.rect.x = PADDLE_START_X

        self.ball.reset()

    def increase_difficulty(self):
        """
        Increase the difficulty of the game
        """
        self.current_opponent_speed += self.speed_increase_per_level
        self.paddleB.speed = self.current_opponent_speed
        self._reset_positions()

    def get_observation(self):
        """
        Return the current observation from the game state.
        """
        return np.array(
            [
                self.get_paddle_position(self.paddleA.rect.y),
                self.get_paddle_position(self.paddleB.rect.y),
                (self.ball.rect.center[0] / self.screen_width - 0.5) * 2,
                (self.ball.rect.center[1] / self.screen_height - 0.5) * 2,
                self.ball.velocity[0] / self.screen_width,
                self.ball.velocity[1] / self.screen_height,
            ],
            dtype=np.float32,
        )

    def get_reward(self):
        """
        Return the reward from the current game state.
        """
        return self.reward_function()

    def _classic_reward(self):
        # Reward the agent if they are the winner
        if self.winner == "A":
            return 1 * self.level
        elif self.winner == "B":
            return -1
        return 0



from ..objects.basic_asset import BasicAsset

SPACE_BETWEEN_SWITCH_AND_GATE = 6


def get_gate_from_switch(switch_group):
    return switch_group + SPACE_BETWEEN_SWITCH_AND_GATE


def get_switch_from_gate(switch_group):
    return switch_group - SPACE_BETWEEN_SWITCH_AND_GATE


class Gate(BasicAsset):
    def __init__(self, sprite, size, random_rotation, group, env):
        super().__init__(sprite, size, random_rotation, group, env)
        self.use_collider = True

    def set_gate_active(self, switch_group):
        active = self.group != get_gate_from_switch(switch_group)
        self.use_collider = active
        alpha = 255 if active else 50
        self.image.set_alpha(alpha)



import pygame
from sai_pygame.utils.animation import AnimationBase

from .map import TILE_SIZE


def get_player_tile(rect):
    tile_x = rect.center[0] // TILE_SIZE
    tile_y = rect.center[1] // TILE_SIZE
    return tile_y, tile_x


class Player(AnimationBase):
    def __init__(
        self,
        all_sprites,
        sprite_group,
        size,
        position,
        screen_height,
        speed,
        wall_colliders,
        switch_colliders,
        door_colliders,
        floor,
        goal_colliders,
        game,
    ):
        super().__init__(all_sprites[sprite_group], position, size, 2, [12, 12])

        self.direction_to_sizes = {
            "down": {"aspect": (83, 95), "base": size * 0.95},
            "side": {"aspect": (98, 70), "base": size * 0.75},
            "idle": {"aspect": (91, 83), "base": size * 0.85},
            "up": {"aspect": (83, 105), "base": size},
        }

        self.rect = self.image.get_rect()
        self.collider = self.rect.copy()
        self.all_sprites = all_sprites
        self.animation_direction = {"key": sprite_group, "invert": False}

        self.screen_height = screen_height
        self.speed = speed
        self.wall_colliders = wall_colliders
        self.switch_colliders = switch_colliders
        self.door_colliders = door_colliders
        self.floor = floor
        self.goal_colliders = goal_colliders
        self.game = game

        self.on_switch = False
        self.on_goal = False
        self.moving = False
        self.prev_moving = True
        self.prev_horizontal_invert = False

    def update_position(self, coordinates):
        self.rect.x = coordinates[0]
        self.rect.y = coordinates[1]
        self.collider.x = coordinates[0]
        self.collider.y = coordinates[1]

    def move(self, horizonalAxis, verticleAxis):
        dir = pygame.math.Vector2(horizonalAxis, verticleAxis)

        self.moving = True
        if horizonalAxis != 0:
            self.prev_horizontal_invert = horizonalAxis == -1
            self.animation_direction = {"key": "side", "invert": horizonalAxis == -1}
        elif verticleAxis == 1:
            self.animation_direction = {"key": "down", "invert": False}
        elif verticleAxis == -1:
            self.animation_direction = {"key": "up", "invert": False}
        else:
            self.moving = False

        # TODO: smooth acceleration
        if dir.length() != 0:
            dir.scale_to_length(self.speed)
            self.collider.x = self.rect.x + dir.x
            self.collider.y = self.rect.y + dir.y
            gate_collision = self.collider.collidelist(self.door_colliders)
            if (
                (gate_collision != -1 and self.game.gates[gate_collision].use_collider)
                or self.collider.collidelist(self.wall_colliders) != -1
                or not self.floor.contains(self.collider)
            ):
                self.collider.x = self.rect.x
                self.collider.y = self.rect.y
                self.game.player_collided(gate_collision != -1)
            else:
                self.rect.x += dir.x
                self.rect.y += dir.y
                switch_collision = self.collider.collidelist(self.switch_colliders)
                if switch_collision != -1:
                    if not self.on_switch:
                        self.game.update_switch(
                            self.game.switches[switch_collision].group
                        )
                        self.on_switch = True
                elif self.on_switch:
                    self.on_switch = False
                goal_collision = self.collider.collidelist(self.goal_colliders)
                if goal_collision != -1:
                    if not self.on_goal:
                        self.on_goal = True
                        self.game.player_reached_goal()
                elif self.on_goal:
                    self.on_goal = False

    def get_new_size(self, direction_key="idle"):
        new_sizes = self.direction_to_sizes[direction_key]
        aspect = new_sizes["aspect"][0] / new_sizes["aspect"][1]
        self.base_size = new_sizes["base"]
        self.invert_axis["x"] = self.animation_direction["invert"]
        return (self.base_size * aspect, self.base_size)

    def update(self):
        if self.moving:
            self.sprites = self.all_sprites[self.animation_direction["key"]]
            self.sizes = self.get_new_size(self.animation_direction["key"])
            super().update(True)
            self.prev_moving = True
        elif self.prev_moving:
            self.image = self.all_sprites["idle"]
            self.animation_direction["invert"] = self.prev_horizontal_invert
            self.sizes = self.get_new_size()
            self.image = pygame.transform.scale(
                self.image, (self.sizes[0], self.sizes[1])
            )
            self.image = pygame.transform.flip(
                self.image, self.invert_axis["x"], self.invert_axis["y"]
            )
            self.prev_moving = False

    def reset(self):
        self.reset_animation()
        self.on_goal = False
        self.on_switch = False
        self.moving = False
        self.prev_moving = True
        self.prev_horizontal_invert = False



from typing import Optional

import pygame
import numpy as np
import gymnasium as gym

from sai_pygame.utils.play import ActionManager


class ArenaXGameBase(gym.Env):
    metadata = {
        "render_modes": ["human", "rgb_array"],
        "engine": "pygame",
    }

    def __init__(
        self,
        width: int,
        height: int,
        framerate: int,
        render_mode: str = "rgb_array",
        game_name: str = "SAI - Minigame",
        action_mapping: dict = {},
        seed: Optional[int] = None,
        index=0
    ):
        assert render_mode is None or render_mode in self.metadata["render_modes"]
        self.framerate = framerate
        self.game_name = game_name

        self.action_manager = ActionManager(action_mapping)

        self.time = 0
        self.frame = 1
        self.done = False
        self.truncated = False

        self.screen_width = width
        self.screen_height = height

        self.clock = pygame.time.Clock()
        self.set_render_mode(render_mode)

    ## Custom methods
    def set_render_mode(self, new_render_mode):
        self.close()
        self.render_mode = new_render_mode

        pygame.font.init()
        if self.render_mode == "human":
            pygame.display.init()
            self.screen = pygame.display.set_mode(
                (self.screen_width, self.screen_height)
            )
            pygame.display.set_caption(self.game_name)
        else:
            self.screen = pygame.Surface((self.screen_width, self.screen_height))

    def get_human_action(self, keys_pressed):
        return self.action_manager.get_action(keys_pressed)

    ## Gym methods
    def reset(self, seed=None, options=None):
        """
        Reset the game to its initial state.
        """
        super().reset(seed=seed)

        self.time = 0
        self.frame = 1
        self.done = False
        self.truncated = False

    def step(self):
        """
        Move the game state forward one frame, given an action for the paddle(s).
        """
        # increment frame and update time
        self.frame += 1
        self.time = self.frame / self.framerate

    def render(self):
        """
        Render the game based on the mode.
        """
        if self.render_mode == "human":
            self.clock.tick(self.framerate)
            pygame.event.pump()
            pygame.display.flip()
        elif self.render_mode == "rgb_array":
            return np.copy(
                np.transpose(pygame.surfarray.pixels3d(self.screen), (1, 0, 2))
            )

    def close(self):
        pygame.quit()



import random
import pygame

from sai_pygame.utils.animation import AnimationBase

from .squid_explosion import SquidExplosion


class Squid(AnimationBase):
    def __init__(self, squid_sprites, x, y, w, h, game):
        super().__init__(squid_sprites, (x, y), (w, h), 3, [5, 5, 5])
        self.x = x
        self.y = y
        self.w = w
        self.h = h
        self.xspd = -5
        self.yspd = 0
        self.game = game

        self.image = random.choice(squid_sprites)
        self.image = pygame.transform.scale(self.image, (self.w, self.h))
        self.rect = pygame.Rect(x, y, w, h)

        self.explosion = None

    def update(self):
        self.rect.x = self.rect.x + self.xspd
        self.rect.y = self.rect.y + self.yspd
        if self.rect.x < 0 - self.w:
            self.game.squids_to_remove.add(self)

        return super().update(loop_bool=True)

    def explode(self, explosion_sprites, size):
        self.explosion = SquidExplosion(
            explosion_sprites,
            (self.rect.center[0], self.rect.center[1]),
            size,
            self.game,
        )
        self.game.all_sprites_list.add(self.explosion)
        self.game.squids_to_remove.add(self)



import pygame

from .input_tracker import InputTracker


class ActionManager:
    def __init__(self, action_bindings: dict):
        """
        Initialize with a dictionary of action bindings.
        action_bindings: dict where each key is an action name and value is a tuple of (pygame_key, action_value)
        Example: {
            "move_up": (pygame.K_w, 1),
            "shoot": (pygame.K_SPACE, 9),
            ...
        }
        """
        self.action_bindings = action_bindings
        self.input_tracker = InputTracker(
            [key for key, _ in self.action_bindings.values()]
        )

    def print_input_mapping(self):
        print("\nInput Mappings\n")
        for action, (key, _) in self.action_bindings.items():
            print("{}: {}".format(action, pygame.key.name(key).upper()))

    def get_action(self, keys_pressed):
        """
        Returns action based on keys pressed using the configured bindings.
        Supports both additive and override actions based on the binding configuration.
        """
        self.input_tracker.press(keys_pressed)
        action = 0

        # Process regular key presses
        for action_name, (key, value) in self.action_bindings.items():
            if keys_pressed[key]:
                if isinstance(value, (int, float)):
                    # Additive action
                    action += value
                elif callable(value):
                    # Custom action function
                    action = value(action)
                else:
                    # Override action
                    action = value

        return action



import math
import pygame

from sai_pygame.utils.animation import AnimationBase
from ..constants.config import SHIP_SIZE


class ShipExhaust(AnimationBase):
    def __init__(self, ship, sprites, position, size=(95, 10)):
        super().__init__(sprites, position, size, 2, [5, 5])
        x_offset = (size[0] * 0.25) + SHIP_SIZE * 0.5
        y_offset = SHIP_SIZE * 0.4

        # triangle offset between center of ship and center of flame
        # pos_offset is the hypotenuse of this triangle
        # pos_angle_offset is the angle of this hypotenuse
        self.pos_offset = math.sqrt(x_offset**2 + y_offset**2)
        self.pos_angle_offset = math.atan(y_offset / x_offset)

    def sync_with_ship(self, ship):
        rotated_image = pygame.transform.rotate(self.base_image, ship.angle)
        self.rect = rotated_image.get_rect(
            center=self.image.get_rect(
                center=(self.sizes[0] / 2, self.sizes[1] / 2)
            ).center
        )
        self.image = rotated_image

        x = ship.center_pos[0] - self.pos_offset * math.cos(
            math.radians(ship.angle) + self.pos_angle_offset
        )
        y = ship.center_pos[1] + self.pos_offset * math.sin(
            math.radians(ship.angle) + self.pos_angle_offset
        )
        self.rect.center = (x, y)

    def update(self, ship):
        kill_bool = super().update(loop_bool=True)
        self.sync_with_ship(ship)

        return kill_bool



import os
import pygame

from ..assets import GAME_ASSETS_BASE


class Rainbow(pygame.sprite.Sprite):
    def __init__(self, x, y, image, lane, game, isStar=False):
        super().__init__()
        self.x = x
        self.y = y
        self.w = image.get_width()
        self.h = image.get_height()
        self.lane = lane
        self.game = game
        self.image = image
        self.rect = pygame.Rect(x, y, self.w, self.h)
        if not isStar:
            for i in range(3):
                startype = game.np_random.integers(1, 4)
                starx = game.np_random.integers(self.x + 10, self.x + self.w - 10)
                stary = game.np_random.integers(self.y + 15, self.y + self.h - 15)
                star = pygame.image.load(
                    os.path.join(
                        GAME_ASSETS_BASE, "rainbow/stars " + str(startype) + ".png"
                    )
                )
                starsprite = Rainbow(starx, stary, star, lane, game, True)
                game.all_sprites_list.add(starsprite)

    def update(self):
        self.y += self.game.speeds[self.lane]
        self.rect.x = self.x
        self.rect.y = self.y
        if self.rect.y > self.game.screen_height:
            self.game.all_sprites_list.remove(self)



import pygame
import numpy as np

from gymnasium.spaces import Discrete, Box

from sai_pygame.utils.env import ArenaXGameBase
from sai_pygame.utils.sprites import load_sprites_from_separate_files
from sai_pygame.utils.colors import BLACK

from .objects.drawable import Drawable
from .objects.basic_asset import BasicAsset

from .entities.player import Player, get_player_tile
from .entities.map import (
    Map,
    TILE_SIZE,
    grids,
    switch_assets,
    door_assets,
    get_channel_value,
    channel_reverse_mapping,
)

from .entities.map import X, SA, SE, G, DA, DE, P1, P2, channel_names
from .entities.gate import Gate, get_gate_from_switch, get_switch_from_gate

from .assets import GAME_ASSETS_BASE

# Game Constants
PLAYER_SIZE = 40
PLAYER_SPEED = 5
PLAYER_ONE_COLOR = (255, 255, 255)
PLAYER_TWO_COLOR = (211, 41, 207)
BACKGROUND_COLOR = (213, 213, 213)
GOAL_COLOR = (0, 255, 245)

valid_grids = {"small": 3, "med": 4, "large": 0}

FRAMERATE = 60
SCREEN_WIDTH = 800
SCREEN_HEIGHT = 900

ACTION_MAPPING = {
    "up": (pygame.K_w, 1),
    "down": (pygame.K_s, 2),
    "left": (pygame.K_a, 3),
    "right": (pygame.K_d, 4),
    "switch": (pygame.K_j, 9),
}


class CoopPuzzleEnv(ArenaXGameBase):
    metadata = {
        "render_modes": ["human", "rgb_array"],
        "render_fps": FRAMERATE,
        "width": SCREEN_WIDTH,
        "height": SCREEN_HEIGHT,
        "engine": "pygame",
        "reward_functions": [
            "classic",
        ],
    }

    def __init__(
        self,
        grid_size="large",
        render_mode="rgb_array",
        framerate=FRAMERATE,
        screen_width=SCREEN_WIDTH,  # Use this to resize screen
        screen_height=SCREEN_HEIGHT,  # Use this to resize screen
        seed=None,
        reward_function="classic",
        **kwargs,
    ):
        if grid_size not in valid_grids:
            raise ValueError("{} is not a valid grid size".format(grid_size))
        self.grid_index = valid_grids[grid_size]

        super().__init__(
            width=TILE_SIZE * len(grids[self.grid_index][0]),
            height=TILE_SIZE * len(grids[self.grid_index]),
            framerate=framerate,
            render_mode=render_mode,
            game_name="Co-op Puzzle - ArenaX Labs",
            action_mapping=ACTION_MAPPING,
            seed=seed,
            **kwargs,
        )

        # Terminal states
        self.TIME_LIMIT = 60 * 2  # 5 minutes

        # Sprite group
        self.all_sprites_list = pygame.sprite.Group()
        self.wall_colliders = []
        self.switch_colliders = []
        self.door_colliders = []
        self.goal_colliders = []
        self.shrub_shadows = {}
        self.switches = []
        self.gates = []
        self.active_switch = 0
        self.won_game = False
        self.player_on_goal = False
        self.player_moved = False
        self.player_switch = False
        self.player_collided_wall = False
        self.player_hit_switch = False
        self.player_hit_different_switch = False

        self.observation_channels = {key: [] for key in channel_names}

        # Environment Assets
        self.flower_sprites = self.get_colored_sprites("flower")
        self.shrub_sprites = self.get_colored_sprites("shrub", 3)
        self.env_sprites = self.get_environment_sprites()

        # Build Map
        self.screen.fill(BLACK)
        self.build_map(self.grid_index)

        # Players
        self.player_one = Player(
            self.get_player_sprites(1),
            "down",
            PLAYER_SIZE,
            self.player_one_start,
            self.screen_height,
            PLAYER_SPEED,
            self.wall_colliders,
            self.switch_colliders,
            self.door_colliders,
            self.floor.collider,
            self.goal_colliders,
            self,
        )
        self.player_two = Player(
            self.get_player_sprites(2),
            "up",
            PLAYER_SIZE,
            self.player_two_start,
            self.screen_height,
            PLAYER_SPEED,
            self.wall_colliders,
            self.switch_colliders,
            self.door_colliders,
            self.floor.collider,
            self.goal_colliders,
            self,
        )
        self.players = [self.player_one, self.player_two]
        self.active_player = 0
        self.end_turn_bool = False
        self.all_sprites_list.add(self.player_one, self.player_two)

        # reset game
        self.reset()

        # define action and observation space
        self.action_space = Discrete(10)
        self.observation_space = Box(
            low=-1,
            high=1,
            shape=(self.map.height, self.map.width, 13),
            dtype=np.float32,
        )

        # reward function
        self.reward_functions = {
            "classic": self._classic_reward,
        }
        self.reward_function = self.reward_functions[reward_function]
        assert self.reward_function is not None, "Invalid reward function"

        # get initial state
        self.init_obs = self.get_observation()

    def get_player_sprites(self, id):
        all_sprites = {
            "idle": pygame.image.load(
                "{}/player{}/stopped.png".format(GAME_ASSETS_BASE, id)
            )
        }
        for direction in ["down", "side", "up"]:
            all_sprites[direction] = load_sprites_from_separate_files(
                "{}/player{}/run-{}".format(GAME_ASSETS_BASE, id, direction), 2
            )
        return all_sprites

    def get_colored_sprites(self, name, num_variations=1):
        sprites = {}
        for color in ["blue", "orange", "pink", "white", "red"]:
            if num_variations == 1:
                asset_key = "{}/{}s/{}-{}.png".format(
                    GAME_ASSETS_BASE, name, name, color
                )
                sprites[color] = pygame.image.load(asset_key)
            else:
                for i in range(num_variations):
                    variation = "{}-{}--{}".format(name, color, i + 1)
                    asset_key = "{}/{}s/{}.png".format(
                        GAME_ASSETS_BASE, name, variation
                    )
                    sprites[variation] = pygame.image.load(asset_key)
        return sprites

    def get_environment_sprites(self):
        return {
            "grass--1": pygame.image.load(
                "{}/environment/grass--1.png".format(GAME_ASSETS_BASE)
            ),
            "grass--2": pygame.image.load(
                "{}/environment/grass--2.png".format(GAME_ASSETS_BASE)
            ),
            "grass--3": pygame.image.load(
                "{}/environment/grass--3.png".format(GAME_ASSETS_BASE)
            ),
            "grass--4": pygame.image.load(
                "{}/environment/grass--4.png".format(GAME_ASSETS_BASE)
            ),
            "grass--5": pygame.image.load(
                "{}/environment/grass--5.png".format(GAME_ASSETS_BASE)
            ),
            "grass-shadow": pygame.image.load(
                "{}/environment/grass-shadow.png".format(GAME_ASSETS_BASE)
            ),
            "wall": pygame.image.load(
                "{}/environment/wall.png".format(GAME_ASSETS_BASE)
            ),
            "cheese": pygame.image.load(
                "{}/environment/cheese.png".format(GAME_ASSETS_BASE)
            ),
        }

    def reset(self, **kwargs):
        """
        Reset the game to its initial state.
        """
        super().reset(**kwargs)
        self.active_switch = -1
        self.won_game = False
        self.active_player = 0
        self.end_turn_bool = False
        self.player_on_goal = False
        self.player_moved = False
        self.player_switch = False
        self.player_collided_wall = False
        self.player_hit_switch = False
        self.player_hit_different_switch = False

        for gate in self.gates:
            gate.set_gate_active(self.active_switch)

        self.player_one.update_position(self.player_one_start)
        self.player_two.update_position(self.player_two_start)
        self.player_one.reset()
        self.player_two.reset()

        observation = self.get_observation()
        info = {"timestep": self.frame}

        return observation, info

    def step(self, action):
        """
        Perform one step of the game, then extract the observation.
        """
        self.update(action)
        super().step()
        observation = self.get_observation()

        done = self.done
        truncated = done or self.TIME_LIMIT <= self.time  # Truncate at 2 min (at 60fps)
        reward = self.get_reward()

        info = {"timestep": self.frame}

        return observation, reward, done, truncated, info

    def update(self, action):
        """
        Move the game state forward one frame, given an action for the paddle(s).
        """

        # Update all sprites
        self.all_sprites_list.update()

        # Movement
        dir = self.get_direction(action)
        self.players[self.active_player].move(dir[1], dir[0])
        if action != 0 and action != 9:
            self.player_moved = True
            self.update_position_channel()

        if action == 9:
            if not self.end_turn_bool:
                self.end_turn()
                self.observation_channels["position"][
                    self.observation_channels["position"] != 0
                ] *= -1
                self.end_turn_bool = True
        else:
            self.end_turn_bool = False

        if self.won_game:
            self.won_game = False

        if self.TIME_LIMIT - self.time <= 0:
            self.done = True

    def render(self):
        """
        Render the game based on the mode.
        """
        # draw everything to the screen
        # self.screen.fill(BACKGROUND_COLOR)
        self.all_sprites_list.draw(self.screen)

        for player in self.players:
            player.update()

        return super().render()

    def get_observation(self):
        """
        Return the current observation from the game state.
        """
        return np.stack(
            [self.observation_channels[c] for c in channel_names], axis=2
        ).astype(np.float32)

    def get_reward(self):
        """
        Return the reward from the current game state.
        """
        return self.reward_function()

    def _classic_reward(self):
        reward = 0
        if self.done:
            reward = 1

        self.player_on_goal = False
        self.player_moved = False
        self.player_switch = False
        self.player_collided_wall = False
        self.player_hit_switch = False
        self.player_hit_different_switch = False

        return reward

    def end_turn(self):
        self.player_switch = True
        self.active_player += 1
        if self.active_player == len(self.players):
            self.active_player = 0

    def add_asset(self, AssetClass, sprite, x, y, random_rotation, group):
        tile = AssetClass(sprite, self.map.tile_size, random_rotation, group, self)
        tile.update_position(self.map.get_world_coordinates(x, y))
        self.all_sprites_list.add(tile)
        return tile

    def add_basic_asset(self, sprite, x, y, random_rotation=False, group=None):
        return self.add_asset(BasicAsset, sprite, x, y, random_rotation, group)

    def get_asset_key(self, group, num_variations):
        variation = self.np_random.integers(1, num_variations)
        return "{}--{}".format(group, variation)

    def add_array_to_channels(self):
        for channel_name in channel_names:
            self.observation_channels[channel_name].append([])

    def add_value_to_channels(self, output):
        for channel_name in channel_names:
            self.observation_channels[channel_name][-1].append(
                get_channel_value(channel_name, output, -1)
            )

    def format_channels(self):
        for channel_name in channel_names:
            self.observation_channels[channel_name] = np.flip(
                self.observation_channels[channel_name]
            )

    def update_door_switch_channel(self, channel_name, turned_on):
        value = -1
        if turned_on:
            value = 1
        self.observation_channels[channel_name] = (
            np.abs(self.observation_channels[channel_name]) * value
        )

    def update_position_channel(self):
        current_player_tiles = get_player_tile(self.players[self.active_player].rect)
        other_player_tiles = get_player_tile(
            self.players[(self.active_player + 1) % 2].rect
        )
        self.observation_channels["position"] = np.zeros(
            self.observation_channels["position"].shape
        )
        self.observation_channels["position"][current_player_tiles[0]][
            current_player_tiles[1]
        ] = 1
        self.observation_channels["position"][other_player_tiles[0]][
            other_player_tiles[1]
        ] = -1

    def build_map(self, map_index):
        self.map = Map(map_index)
        self.floor = Drawable(
            self.map.width * self.map.tile_size,
            self.map.height * self.map.tile_size,
            BACKGROUND_COLOR,
        )
        self.grass = []
        self.all_sprites_list.add(self.floor)
        for y in reversed(range(self.map.height)):
            self.add_array_to_channels()
            for x in reversed(range(self.map.width)):
                self.add_basic_asset(
                    self.env_sprites[self.get_asset_key("grass", 5)], x, y, True
                )

                output = self.map.grid[y][x]
                self.add_value_to_channels(output)
                if output >= SA and output <= SE:
                    switch = self.add_basic_asset(
                        self.flower_sprites[switch_assets[output]], x, y, False, output
                    )
                    self.switch_colliders.append(switch.collider)
                    self.switches.append(switch)
                elif output >= DA and output <= DE:
                    shrub_shadow = self.add_basic_asset(
                        self.env_sprites["grass-shadow"], x, y + 0.2, False
                    )
                    if output not in self.shrub_shadows:
                        self.shrub_shadows[output] = []
                    self.shrub_shadows[output].append(shrub_shadow)
                    gate = self.add_asset(
                        Gate,
                        self.shrub_sprites[
                            self.get_asset_key(
                                "shrub-{}".format(door_assets[output]), 3
                            )
                        ],
                        x,
                        y,
                        True,
                        output,
                    )
                    self.door_colliders.append(gate.collider)
                    self.gates.append(gate)
                elif output == X:
                    if self.map.grid[y + 1][x] != X:
                        self.add_basic_asset(
                            self.env_sprites["grass-shadow"], x, y + 0.2, False
                        )
                    wall = self.add_basic_asset(self.env_sprites["wall"], x, y, True)
                    self.wall_colliders.append(wall.collider)
                elif output == P1:
                    self.player_one_start = self.map.get_world_coordinates(x, y)
                elif output == P2:
                    self.player_two_start = self.map.get_world_coordinates(x, y)
                elif output == G:
                    goal = self.add_basic_asset(self.env_sprites["cheese"], x, y, False)
                    self.goal_colliders.append(goal.collider)
        self.format_channels()

    def update_switch(self, switch_group):
        self.player_hit_switch = True
        target_gate = get_gate_from_switch(switch_group)
        if switch_group == self.active_switch:
            return

        gates_updated = {}
        for gate in self.gates:
            gate.set_gate_active(switch_group)
            gate_name = channel_reverse_mapping[gate.group]
            if gate_name not in gates_updated:
                current_switch = get_switch_from_gate(gate.group)
                self.update_door_switch_channel(
                    channel_reverse_mapping[current_switch],
                    current_switch == switch_group,
                )
                self.update_door_switch_channel(
                    channel_reverse_mapping[gate.group], gate.group == target_gate
                )
                gates_updated[gate_name] = True

        for group, shadows in self.shrub_shadows.items():
            if group == get_gate_from_switch(switch_group):
                alpha = 0
            else:
                alpha = 255
            for shadow in shadows:
                shadow.image.set_alpha(alpha)

        self.player_hit_different_switch = True
        self.active_switch = switch_group

    def player_reached_goal(self):
        self.player_on_goal = True
        if self.player_one.on_goal and self.player_two.on_goal:
            self.won_game = True
            self.done = True

    def player_collided(self, gate):
        if not gate:
            self.player_collided_wall = True

    def get_direction(self, action):
        if action == 1:
            return [-1, -1]
        elif action == 2:
            return [-1, 1]
        elif action == 3:
            return [-1, 0]
        elif action == 4:
            return [1, -1]
        elif action == 5:
            return [1, 1]
        elif action == 6:
            return [1, 0]
        elif action == 7:
            return [0, -1]
        elif action == 8:
            return [0, 1]
        else:
            return [0, 0]



class InputTracker:
    def __init__(self, keys_to_track):
        self.keys_to_track = keys_to_track
        self.pressed = {}
        self.held = {}
        self.reset()

    def press(self, keys_pressed):
        for key in self.keys_to_track:
            self.pressed[key] = keys_pressed[key]
            if keys_pressed[key]:
                self.held[key] += 1
            else:
                self.held[key] = 0

    def reset(self):
        for key in self.keys_to_track:
            self.pressed[key] = False
            self.held[key] = 0



import pygame
import gymnasium as gym
from .paddle import Paddle
BLACK = (0, 0, 0)


class Ball(pygame.sprite.Sprite):
    def __init__(self, env: gym.Env, color, diameter, screen_width, screen_height, max_speed=7):
        super().__init__()

        self.screen_width = screen_width
        self.screen_height = screen_height
        self.diameter = diameter
        self.env = env

        self.max_speed = max_speed

        self.image = pygame.Surface([self.diameter, self.diameter])
        self.image.fill(BLACK)
        self.image.set_colorkey(BLACK)

        pygame.draw.rect(self.image, color, [0, 0, self.diameter, self.diameter])

        self.rect = self.image.get_rect()
        self.velocity: list[float] = [0.0, 0.0]
        self.reset()

    def reset(self):
        direction = self.env.np_random.choice([-1, 1])
        self.velocity: list[float] = [
            direction * self.env.np_random.integers(3, self.max_speed),
            self.env.np_random.integers(-3, 3),
        ]
        # Center the ball
        self.rect.x = self.screen_width // 2 - self.diameter // 2
        self.rect.y = self.screen_height // 2 - self.diameter // 2

    def update(self):
        self.rect.x += int(self.velocity[0])
        self.rect.y += int(self.velocity[1])

    def bounce(self, paddle: Paddle):
        self.velocity[0] = max(-self.max_speed, min(self.max_speed, self.velocity[0] * -1.1) )
    
        # Calculate where ball hit the paddle (0.0 = top, 1.0 = bottom)
        ball_center_y = self.rect.centery
        paddle_center_y = paddle.rect.centery
        paddle_half_height = paddle.height / 2
        
        # Hit position: -1.0 (top) to 1.0 (bottom)
        hit_position = (ball_center_y - paddle_center_y) / paddle_half_height
        hit_position = max(-1.0, min(1.0, hit_position))

        # Add random valruable to prevent constant left-right movement
        if abs(hit_position) < 0.1:
            hit_position += self.env.np_random.uniform(-0.1, 0.1)
        
        # Set vertical velocity based on hit position
        self.velocity[1] = hit_position * self.max_speed



import pygame


class AnimationBase(pygame.sprite.Sprite):
    def __init__(self, sprites, position, size, num_frames, frame_frequencies=[]):
        super().__init__()
        self.sprites = sprites
        self.position = position
        self.num_frames = num_frames
        self.base_size = size
        if isinstance(size, int):
            self.sizes = (size, size)
        elif isinstance(size, tuple) and len(size) == 2:
            self.sizes = size
        else:
            raise ValueError("Size must be an int or a tuple of two ints.")
        self.frame = 0
        self.time_on_frame = 0
        self.invert_axis = {"x": False, "y": False}
        self.rotation = None
        self.change_sprite()
        self.empty_surface = pygame.Surface([0, 0])
        self.rect = self.image.get_rect(center=(position[0], position[1]))
        if len(frame_frequencies) == 0:
            self.frame_frequencies = [3 for _ in range(num_frames)]
        else:
            self.frame_frequencies = frame_frequencies

    def update_frame(self):
        self.time_on_frame += 1
        if self.frame >= self.num_frames:
            return True
        if self.time_on_frame >= self.frame_frequencies[self.frame]:
            self.frame += 1
            self.time_on_frame = 0
        return self.frame == self.num_frames

    def change_sprite(self):
        image = self.sprites[self.frame]
        image = pygame.transform.scale(image, (self.sizes[0], self.sizes[1]))
        image = pygame.transform.flip(
            image, self.invert_axis["x"], self.invert_axis["y"]
        )
        if self.rotation is not None:
            image = pygame.transform.rotate(image, self.rotation)
        self.base_image = image
        self.image = image

    def reset_animation(self):
        self.frame = 0
        self.time_on_frame = 0

    def update(self, loop_bool=False, stay_on_last_frame=False):
        kill_bool = self.update_frame()
        if not kill_bool:
            self.change_sprite()
        elif not stay_on_last_frame:
            if loop_bool:
                self.reset_animation()
            else:
                self.image = self.empty_surface
        return kill_bool



import pygame

BLACK = (0, 0, 0)


class Paddle(pygame.sprite.Sprite):
    def __init__(self, color, width, height, screen_height, speed):
        super().__init__()

        self.screen_height = screen_height
        self.width = width
        self.height = height
        self.speed = speed

        self.image = pygame.Surface([self.width, self.height])
        self.image.fill(BLACK)
        self.image.set_colorkey(BLACK)

        pygame.draw.rect(self.image, color, [0, 0, self.width, self.height])

        self.rect = self.image.get_rect()

    def moveUp(self):
        if self.rect.y > 0:  # Can move until top edge
            self.rect.y = max(0, self.rect.y - self.speed)


    def moveDown(self):
        if self.rect.y + self.height < self.screen_height:  # Can move until bottom edge
            self.rect.y = min(self.screen_height - self.height, self.rect.y + self.speed)



import pygame

from .map import switch_assets


class Switch(pygame.sprite.Sprite):
    def __init__(self, sprites, group, size):
        super().__init__()
        self.group = group
        self.image = pygame.transform.scale(sprites[switch_assets[group]], (size, size))
        self.rect = self.image.get_rect()
        self.collider = self.rect.copy()

    def update_position(self, coordinates):
        self.rect.x = coordinates[0]
        self.rect.y = coordinates[1]
        self.collider.x = coordinates[0]
        self.collider.y = coordinates[1]



import pygame

from sai_pygame.utils.animation import AnimationBase

from ..constants import SCREEN_HEIGHT, SCREEN_WIDTH, SHIP_SIZE


class Ship(AnimationBase):
    def __init__(self, ship_sprites, x, y):
        super().__init__(ship_sprites, (x, y), (SHIP_SIZE, SHIP_SIZE), 3, [5, 5, 5])
        self.x = 0
        self.y = 0
        self.center_pos = (0, 0)
        self.xspd = 0
        self.yspd = 0
        self.speed = 10
        self.angle = 0
        self.lane = 2
        self.previous_lane = 2
        self.moving = None

        self.rect = pygame.Rect(100, 100, SHIP_SIZE, SHIP_SIZE)

    def reset(self):
        self.x = SCREEN_WIDTH / 2 - SHIP_SIZE / 2
        self.y = SCREEN_HEIGHT * 3 / 4 - SHIP_SIZE / 2
        self.lane = 2
        self.previous_lane = 2

    def update(self):
        if self.moving is not None:
            if (
                abs((self.x + SHIP_SIZE / 2) - ((self.moving) * SCREEN_WIDTH / 4))
                <= self.speed
            ):
                self.x = (self.moving) * SCREEN_WIDTH / 4 - SHIP_SIZE / 2
                self.previous_lane = self.lane
                self.lane = self.moving
                self.moving = None
                self.xspd = 0
            else:
                self.xspd = (self.moving - self.lane) * self.speed

        self.x += self.xspd
        self.y += self.yspd
        self.rect.x = int(self.x)
        self.rect.y = int(self.y)
        return super().update(loop_bool=True)

    def move(self, dir):
        if 1 <= self.lane + dir <= 3:
            self.moving = self.lane + dir



from sai_pygame.utils.animation import AnimationBase


class SquidExplosion(AnimationBase):
    def __init__(self, explosion_sprites, position, size, game):
        super().__init__(explosion_sprites, position, size, 3, [5, 5, 5])
        self.game = game

    def update(self):
        kill_bool = super().update(loop_bool=True)

        if kill_bool:
            self.game.all_sprites_list.remove(self)

        return kill_bool



import pygame
import os

from ..assets import GAME_ASSETS_BASE


class Drawable(pygame.sprite.Sprite):
    def __init__(self, width, height, color, image_name=""):
        super().__init__()

        self.width = width
        self.height = height
        self.color = color
        self.use_collider = True

        self.image = pygame.Surface([self.width, self.height])
        pygame.draw.rect(self.image, self.color, [0, 0, self.width, self.height])
        if image_name != "":
            image = pygame.image.load(os.path.join(GAME_ASSETS_BASE, image_name))
            image = pygame.transform.scale(image, (self.width, self.height))
            self.image.blit(image, self.image.get_rect())

        self.rect = self.image.get_rect()
        self.collider = self.rect.copy()

    def update_position(self, coordinates):
        self.rect.x = coordinates[0]
        self.rect.y = coordinates[1]
        self.collider.x = coordinates[0]
        self.collider.y = coordinates[1]



import pygame


class Asteroid(pygame.sprite.Sprite):
    def __init__(self, x, y, w, h, image, lane, game):
        super().__init__()
        self.x = x
        self.y = y
        self.w = w
        self.h = h
        self.game = game
        self.lane = lane
        self.image = image
        self.rect = pygame.Rect(x, y, w, h)
        self.createTime = self.game.time
        self.start = (
            self.createTime
            + self.game.ship.y / self.game.speeds[self.lane] / self.game.framerate
        )
        self.end = (
            self.start + self.h / self.game.speeds[self.lane] / self.game.framerate
        )

    def update(self):
        self.y += self.game.speeds[self.lane]
        self.start = (
            self.createTime
            + self.game.ship.y / self.game.speeds[self.lane] / self.game.framerate
        )
        self.end = (
            self.start + self.h / self.game.speeds[self.lane] / self.game.framerate
        )
        if self.rect.y > self.game.screen_height:
            self.game.all_sprites_list.remove(self)
            self.game.asteroids[self.lane].remove(self)

        self.rect.x = self.x
        self.rect.y = self.y



import pygame
import numpy as np
import os

from gymnasium.spaces import Discrete, Box

from sai_pygame.utils.sprites import load_sprites_from_separate_files
from sai_pygame.utils.env import ArenaXGameBase
from sai_pygame.utils.colors import BLACK

from .entities.ship import Ship
from .entities.asteroid import Asteroid
from .objects.text import Text
from .entities.rainbow import Rainbow
from .assets import GAME_ASSETS_BASE

from .constants.config import (
    FRAMERATE,
    SCREEN_HEIGHT,
    SCREEN_WIDTH,
    SHIP_SIZE,
)

ACTION_MAPPING = {
    "left": (pygame.K_a, 1),
    "right": (pygame.K_d, 2),
}


class SpaceEvadersEnv(ArenaXGameBase):
    metadata = {
        "render_modes": ["human", "rgb_array"],
        "render_fps": FRAMERATE,
        "width": SCREEN_WIDTH,
        "height": SCREEN_HEIGHT,
        "engine": "pygame",
        "reward_functions": [
            "classic",
            "minimum_lane_change",
        ],
    }

    def __init__(
        self,
        render_mode="rgb_array",
        width=SCREEN_WIDTH,
        height=SCREEN_HEIGHT,
        framerate=FRAMERATE,
        seed=None,
        reward_function="classic",
        **kwargs,
    ):
        super().__init__(
            width=width,
            height=height,
            framerate=framerate,
            render_mode=render_mode,
            game_name="Space Evaders - ArenaX Labs",
            action_mapping=ACTION_MAPPING,
            seed=seed,
            **kwargs,
        )
        # initialize game assets
        self.all_sprites_list = pygame.sprite.LayeredUpdates()
        self.speedinterval = self.time
        self.TIME_LIMIT = 60
        self.ASTEROID_INTERVAL = 100
        self.SHIP_SIZE = SHIP_SIZE
        self.NUM_LANES = 3
        self.slowest = 4
        self.fastest = 6
        self.maxspeed = 15
        self.speeds = [
            self.np_random.integers(self.slowest, self.fastest),
            self.np_random.integers(self.slowest, self.fastest),
            self.np_random.integers(self.slowest, self.fastest),
        ]
        ship_sprites = load_sprites_from_separate_files(
            "{}/ship/ship".format(GAME_ASSETS_BASE), 3
        )
        self.time_text = Text(
            self.screen_width / 2, 30, str(round(self.time, 2)), 30, self
        )
        self.rainbows = []
        self.bg = pygame.image.load(os.path.join(GAME_ASSETS_BASE, "background.png"))
        self.rainbow = pygame.image.load(
            os.path.join(GAME_ASSETS_BASE, "rainbow/rainbow.png")
        )
        self.rainbowsec = pygame.image.load(
            os.path.join(GAME_ASSETS_BASE, "rainbow/rainbow section.png")
        )
        self.rainbowsec = pygame.transform.scale(
            self.rainbowsec, (self.rainbowsec.get_width(), self.rainbowsec.get_height())
        )
        self.dead = False
        self.lastrainbows = [None] * self.NUM_LANES
        self.asteroids = [[] for i in range(self.NUM_LANES)]
        self.nextAsteroid = self.np_random.integers(0, self.NUM_LANES)
        self.gridrows = 6
        self.gridcolumns = self.NUM_LANES
        for i in range(self.NUM_LANES):
            x = (i + 1) * (self.screen_width) / 4
            rainbow = Rainbow(
                x - self.rainbowsec.get_width() / 2,
                -self.rainbowsec.get_height(),
                self.rainbowsec,
                i,
                self,
            )
            self.lastrainbows[i] = rainbow
            self.all_sprites_list.add(rainbow)
        x = SCREEN_WIDTH / 2 - SHIP_SIZE / 2
        y = SCREEN_HEIGHT * 3 / 4 - SHIP_SIZE / 2
        self.ship = Ship(ship_sprites, x, y)
        self.all_sprites_list.add(self.ship)

        # reset game
        self.reset()

        # define action and observation space
        self.action_space = Discrete(3)
        self.observation_space = Box(
            low=np.array([-1, 0, 0, 0] + [0] * (self.gridcolumns * self.gridrows)),
            high=np.array([1] * 4 + [1] * (self.gridcolumns * self.gridrows)),
            dtype=np.float32,
        )

        # initialize reward function
        self.reward_functions = {
            "classic": self._classic_reward,
            "minimum_lane_change": self._minimum_lane_change_reward,
        }
        self.reward_function = self.reward_functions[reward_function]
        assert reward_function in self.reward_functions, (
            f"Invalid reward function: {reward_function}"
        )

        # get initial state
        self.init_obs = self.get_observation()

    def reset(self, **kwargs):
        """
        Reset the game to its initial state.
        """
        super().reset(**kwargs)
        # reset the game environment back to the initial state
        self.speeds = [
            self.np_random.integers(self.slowest, self.fastest),
            self.np_random.integers(self.slowest, self.fastest),
            self.np_random.integers(self.slowest, self.fastest),
        ]
        self.all_sprites_list = pygame.sprite.LayeredUpdates()
        self.asteroids = [[] for i in range(self.NUM_LANES)]
        for i in range(self.NUM_LANES):
            x = (i + 1) * (self.screen_width) / 4
            rainbow = Rainbow(
                x - self.rainbowsec.get_width() / 2,
                -self.rainbowsec.get_height(),
                self.rainbowsec,
                i,
                self,
            )
            self.lastrainbows[i] = rainbow
            self.all_sprites_list.add(rainbow)
        for i in range(self.NUM_LANES):
            x = (i + 1) * (self.screen_width) / 4
            r = Rainbow(x - self.rainbow.get_width() / 2, -80, self.rainbow, i, self)
            self.all_sprites_list.add(r)
        self.ship.reset()
        self.all_sprites_list.add(self.ship)
        self.nextAsteroid = self.np_random.integers(0, self.NUM_LANES)
        self.dead = False
        self.speedinterval = self.time

        observation = self.get_observation()

        info = {
            "timestep": self.frame,
            "asteroids": len([a for i in self.asteroids for a in i]),
        }

        return observation, info

    def step(self, action):
        """
        Perform one step of the game, then extract the observation.
        """
        self.update(action)
        super().step()

        observation = self.get_observation()
        done = self.done
        truncated = self.truncated
        reward = self.get_reward()

        info = {
            "timestep": self.frame,
            "asteroids": len([a for i in self.asteroids for a in i]),
        }

        return observation, reward, done, truncated, info

    def update(self, action):
        """
        Move the game state forward one frame, given an action for the paddle(s).
        """
        # game loop goes here
        if self.time - self.speedinterval > 10:
            for i in range(self.NUM_LANES):
                self.speeds[i] = min(self.maxspeed, self.speeds[i] + 1)
            self.speedinterval = self.time

        self.all_sprites_list.update()
        if action == 1:
            self.ship.move(-1)
        if action == 2:
            self.ship.move(1)

        for i in range(self.NUM_LANES):
            if self.lastrainbows[i] is None or self.lastrainbows[i].y > -self.speeds[i]:
                x = (i + 1) * (self.screen_width) / 4
                if self.np_random.integers(1, 9) > 1 and i == self.nextAsteroid:
                    cluster = self.np_random.integers(1, 7)
                    astrImg = pygame.image.load(
                        os.path.join(
                            GAME_ASSETS_BASE,
                            "asteroids/asteroid cluster " + str(cluster) + ".png",
                        )
                    )
                    width = astrImg.get_width()
                    height = astrImg.get_height()
                    asteroid = Asteroid(
                        x - width / 2, -height, width, height, astrImg, i, self
                    )
                    self.asteroids[i].append(asteroid)
                    # make sure generated asteroid won't trap the ship by either keeping middle or both side lanes free
                    if self.checkBlockade():
                        self.asteroids[i].remove(asteroid)
                        rainbow = Rainbow(
                            x - self.rainbowsec.get_width() / 2,
                            -self.rainbowsec.get_height(),
                            self.rainbowsec,
                            i,
                            self,
                        )
                        self.lastrainbows[i] = rainbow
                        self.all_sprites_list.add(rainbow)
                        self.all_sprites_list.move_to_back(rainbow)
                        continue
                    # print([len(i) for i in self.asteroids])
                    self.nextAsteroid = self.np_random.integers(0, self.NUM_LANES)
                    self.all_sprites_list.add(asteroid)
                    self.lastrainbows[i] = asteroid

                else:
                    rainbow = Rainbow(
                        x - self.rainbowsec.get_width() / 2,
                        -self.rainbowsec.get_height(),
                        self.rainbowsec,
                        i,
                        self,
                    )
                    self.lastrainbows[i] = rainbow
                    self.all_sprites_list.add(rainbow)
                    self.all_sprites_list.move_to_back(rainbow)

        self.check_collisions()
        if self.TIME_LIMIT - self.time <= 0:
            self.truncated = True
        self.time_text.update("TIME: " + str(round(self.time, 1)))

    def check_collisions(self):
        for j in self.asteroids:
            for i in j:
                if pygame.Rect(
                    self.ship.rect.x,
                    self.ship.rect.y,
                    self.ship.image.get_width(),
                    self.ship.image.get_height(),
                ).colliderect(i.rect):
                    self.done = True
                    self.dead = True
                    return

    def render(self):
        """
        Render the game based on the mode.
        """
        # draw everything to the screen
        self.screen.fill(BLACK)
        self.screen.blit(self.bg, (0, 0))
        self.all_sprites_list.move_to_front(self.ship)
        self.all_sprites_list.draw(self.screen)
        self.screen.blit(
            self.time_text.image,
            (
                self.time_text.x - self.time_text.image.get_width() / 2,
                self.time_text.y - self.time_text.image.get_height() / 2,
            ),
        )
        return super().render()

    def get_observation(self):
        """
        Return the current observation from the game state.
        """
        # use the game state to output an array in the format of your observation space
        ship_position = ((self.ship.x + SHIP_SIZE / 2) - self.screen_width / 2) / (
            self.screen_width / 4
        )
        ast_casts = []
        # closest asteroids to the ship in each lane
        # for i in self.asteroids:
        #     if len(i) > 0:
        #         ast_casts.append(1 + (i[0].y - (self.ship.y + SHIP_SIZE)) / (self.ship.y + SHIP_SIZE + i[0].h))
        #     else:
        #         ast_casts.append(0)
        self.normalspeeds = [
            (i - self.slowest) / (self.maxspeed - self.slowest) for i in self.speeds
        ]
        columns = self.gridcolumns
        rows = self.gridrows
        intersections = [[0 for i in range(columns)] for j in range(rows)]
        count = 0
        asteroids = [a for i in self.asteroids for a in i]
        # for a in asteroids:
        #     pygame.draw.rect(self.screen, RED, a.rect)

        for i in range(rows):
            for j in range(columns):
                cell = pygame.Rect(
                    SCREEN_WIDTH / columns * j,
                    SCREEN_HEIGHT / rows * i,
                    SCREEN_WIDTH / columns,
                    SCREEN_HEIGHT / rows,
                )
                for a in asteroids:
                    if pygame.Rect.clipline(
                        a.rect,
                        cell.centerx,
                        SCREEN_HEIGHT / rows * i,
                        cell.centerx,
                        SCREEN_HEIGHT / rows * (i + 1),
                    ):
                        intersections[i][j] = 1
                        count += 1
                # pygame.draw.rect(self.screen, RED if intersections[i][j] else BLACK, cell, 3)

        flat_intersections = [j for i in intersections for j in i]
        arr = [ship_position] + self.normalspeeds + flat_intersections
        return np.array(arr, dtype=np.float32)

    def get_reward(self):
        """
        Return the reward from the current game state.
        """
        return self.reward_function()

    def _classic_reward(self):
        if self.dead:
            return -5

        return 0.01

    def _minimum_lane_change_reward(self):
        if self.dead:
            return -5

        lane_changes = abs(self.ship.previous_lane - self.ship.lane)
        lane_change_penalty = -0.1 * lane_changes

        survival_reward = 0.01

        return lane_change_penalty + survival_reward

    def checkBlockade(self):
        delay = SHIP_SIZE * 2 / self.framerate / self.slowest
        for i in range(2):
            # print("i:", i)
            for j in self.asteroids[i]:
                # print("j:", self.asteroids[i].index(j))
                for k in range(i + 1, self.NUM_LANES):
                    # print("k:", i)
                    for l in self.asteroids[k]:
                        # print("l:", self.asteroids[k].index(l))
                        # print(str(round(j.start, 1)),  str(round(j.end, 1)), str(round(l.start, 1)), str(round(l.end, 1)))
                        if not (j.start > l.end + delay or j.end + delay < l.start):
                            if i == 1 or k == 1:
                                return True
        return False



import pygame
import gymnasium as gym

class BasicAsset(pygame.sprite.Sprite):
    def __init__(self, sprite, size, random_rotation, group, env: gym.Env):
        super().__init__()
        self.image = pygame.transform.scale(sprite, (size, size))
        if random_rotation:
            rotation = env.np_random.integers(1, 4) * 90
            self.image = pygame.transform.rotate(self.image, rotation)
        self.rect = self.image.get_rect()
        self.collider = self.rect.copy()
        self.group = group

    def update_position(self, coordinates):
        self.rect.x = coordinates[0]
        self.rect.y = coordinates[1]
        self.collider.x = coordinates[0]
        self.collider.y = coordinates[1]



import pygame
import numpy as np
import os

from sai_pygame.utils.env import ArenaXGameBase
from sai_pygame.utils.sprites import load_sprites_from_separate_files
from sai_pygame.utils.colors import RED, BLACK

from gymnasium.spaces import Discrete, Box

from .assets import GAME_ASSETS_BASE
from .entities.bullet import Bullet
from .entities.ship import Ship
from .entities.squid import Squid
from .objects.text import Text
from .entities.wall import Wall

from .constants.config import (
    FRAMERATE,
    SCREEN_HEIGHT,
    SCREEN_WIDTH,
    SHIP_SIZE,
    BACKGROUND_OPACITY,
    BACKGROUND_SPEED,
    BACKGROUND_HEIGHT,
    BACKGROUND_WIDTH,
    SHIP_MAX_ANGLE,
)


ACTION_MAPPING = {
    "up": (pygame.K_w, 1),
    "down": (pygame.K_s, 2),
    "rotate_cc": (pygame.K_a, 3),
    "rotate_c": (pygame.K_d, 6),
    "shoot": (pygame.K_SPACE, 9),
}


class SquidHuntEnv(ArenaXGameBase):
    metadata = {
        "render_modes": ["human", "rgb_array"],
        "render_fps": FRAMERATE,
        "width": SCREEN_WIDTH,
        "height": SCREEN_HEIGHT,
        "engine": "pygame",
        "reward_functions": [
            "classic",
            "highest_kill_count",
            "precision_shooting",
            "ammo_efficiency",
            "time_survival",
        ],
    }

    def __init__(
        self,
        render_mode="rgb_array",
        width=SCREEN_WIDTH,
        height=SCREEN_HEIGHT,
        framerate=FRAMERATE,
        seed=None,
        reward_function="classic",
        **kwargs,
    ):
        super().__init__(
            width=width,
            height=height,
            framerate=framerate,
            render_mode=render_mode,
            game_name="Squid Hunt - ArenaX Labs",
            action_mapping=ACTION_MAPPING,
            seed=seed,
            **kwargs,
        )
        # initialize game assets
        self.all_sprites_list = pygame.sprite.Group()

        self.TIME_LIMIT = 30
        self.WALL_INTERVAL = 200
        self.SQUID_INTERVAL = 100
        self.AMMO_INTERVAL = 100
        self.MAX_AMMO = 30
        self.INITIAL_AMMO = 10
        # can be "random", "up", or "down"
        self.ENEMYSPAWN = "random"
        self.WALL_SPAWN = True

        self.genY = 0 if self.ENEMYSPAWN == "down" else self.screen_height - 60
        self.SHIP_SIZE = SHIP_SIZE
        self.score = 0
        self.ammo = self.INITIAL_AMMO

        self.city_x_position = 0
        self.city_scroll_speed = BACKGROUND_SPEED

        darkening_surface = pygame.Surface(
            (BACKGROUND_WIDTH, BACKGROUND_HEIGHT), pygame.SRCALPHA
        )
        darkening_surface.fill((0, 0, 0, 255 - BACKGROUND_OPACITY))
        self.city = pygame.image.load(os.path.join(GAME_ASSETS_BASE, "city.jpg"))
        self.city = pygame.transform.scale(
            self.city, (BACKGROUND_WIDTH, BACKGROUND_HEIGHT)
        )
        self.city.blit(darkening_surface, (0, 0))

        self.walls_to_remove = set()
        self.squids_to_remove = set()
        self.bullets_to_remove = set()

        exhaust_sprites = load_sprites_from_separate_files(
            "{}/ship-exhaust/ship-exhaust".format(GAME_ASSETS_BASE), 2
        )
        self.ship = Ship(exhaust_sprites)
        self.all_sprites_list.add(self.ship)

        self.time_text = Text(
            self.screen_width / 2,
            30,
            str(round(self.TIME_LIMIT - self.time, 2)),
            30,
            self,
        )
        self.ammo_text = Text(
            self.screen_width / 3,
            self.screen_height - 30,
            "BULLETS: " + str(self.ammo),
            30,
            self,
        )
        self.score_text = Text(
            self.screen_width / 3 * 2,
            self.screen_height - 30,
            "SCORE: " + str(self.score),
            30,
            self,
        )

        self.wall_sprites = load_sprites_from_separate_files(
            "{}/wall/wall".format(GAME_ASSETS_BASE), 3
        )
        self.projectile_sprites = load_sprites_from_separate_files(
            "{}/projectile/projectile".format(GAME_ASSETS_BASE), 3
        )
        self.squid_sprites = {
            "ORANGE": load_sprites_from_separate_files(
                "{}/squids/squid--orange-".format(GAME_ASSETS_BASE), 3
            ),
            "PINK": load_sprites_from_separate_files(
                "{}/squids/squid--pink-".format(GAME_ASSETS_BASE), 3
            ),
        }
        self.explosion_sprites = load_sprites_from_separate_files(
            "{}/explosion/explosion".format(GAME_ASSETS_BASE), 3
        )

        self.killed = False
        self.shot = False
        self.dead = False
        self.prevY = self.ship.y
        self.shot_action = False
        self.closest_hazard = 9999

        # reset game
        self.reset()

        # define action and observation space
        self.action_space = Discrete(10)
        self.observation_space = Box(
            low=np.array([-1] * 23), high=np.array([1] * 23), dtype=np.float32
        )

        # initialize reward function
        self.reward_functions = {
            "classic": self._classic_reward,
            "highest_kill_count": self._highest_kill_count_reward,
            "precision_shooting": self._precision_shooting_reward,
            "ammo_efficiency": self._ammo_efficiency_reward,
            "time_survival": self._time_survival_reward,
        }
        self.reward_function = self.reward_functions[reward_function]
        assert self.reward_function is not None, (
            f"Invalid reward function: {reward_function}"
        )

        # get initial state
        self.init_obs = self.get_observation()

    def reset(self, **kwargs):
        """
        Reset the game to its initial state.
        """
        super().reset(**kwargs)
        # reset the game environment back to the initial state
        self.genY = 0 if self.ENEMYSPAWN == "down" else self.screen_height - 60
        self.walls = set()
        self.bullets = set()
        self.squids = set()

        self.all_sprites_list = pygame.sprite.Group()
        self.all_sprites_list.add(self.ship)
        self.ship.reset()

        self.ammo = self.INITIAL_AMMO
        self.score = 0
        self.killed = False
        self.shot = False
        self.shot_action = False
        self.dead = False
        self.combo_streak = 0
        self.closest_hazard = self.get_nearest_hazard()
        self.prevY = self.ship.y

        observation = self.get_observation()
        info = self.get_info()
        return observation, info

    def get_reward(self):
        """
        Return the reward from the current game state.
        """
        return self.reward_function()

    def get_info(self):
        """
        Return the info from the current game state.
        """
        info = {
            "timestep": self.frame,
            "time": self.time,
            "num_walls": len(self.walls),
            "num_squids": len(self.squids),
            "num_bullets": len(self.bullets),
            "shotBullet": self.shot,
            "failedShot": self.shot_action,
            "bullets": len(self.bullets),
            "squids": len(self.squids),
            "approaching_danger": self.closest_hazard > self.get_nearest_hazard(),
            "approaching_edge": (
                self.screen_height / 5 > self.prevY > self.ship.y
                or self.screen_height * 0.8 - self.ship.image.get_height()
                < self.prevY
                < self.ship.y
            ),
        }

        self.closest_hazard = self.get_nearest_hazard()
        self.prevY = self.ship.y
        return info

    def render(self):
        """
        Render the game based on the mode.
        """
        # draw everything to the screen
        self.screen.fill(BLACK)
        self.screen.blit(self.city, (self.city_x_position, -100))
        if self.city_x_position <= -self.screen_width + self.screen_width:
            self.screen.blit(
                self.city, (self.city_x_position + self.screen_width, -100)
            )
        self.all_sprites_list.draw(self.screen)
        self.screen.blit(self.ship.exhaust.image, self.ship.exhaust.rect)
        self.screen.blit(
            self.ammo_text.image,
            (
                self.ammo_text.x - self.ammo_text.image.get_width() / 2,
                self.ammo_text.y - self.ammo_text.image.get_height() / 2,
            ),
        )
        self.screen.blit(
            self.time_text.image,
            (
                self.time_text.x - self.time_text.image.get_width() / 2,
                self.time_text.y - self.time_text.image.get_height() / 2,
            ),
        )
        self.screen.blit(
            self.score_text.image,
            (
                self.score_text.x - self.score_text.image.get_width() / 2,
                self.score_text.y - self.score_text.image.get_height() / 2,
            ),
        )

        return super().render()

    def step(self, action):
        """
        Move the game state forward one frame, given an action for the paddle(s).
        """
        self.update(action)
        super().step()

        observation = self.get_observation()
        done = self.done
        truncated = done
        reward = self.get_reward()
        info = self.get_info()

        return observation, reward, done, truncated, info

    def update(self, action):
        """
        Update the game state based on the action.
        """
        self.killed = False
        self.shot = False
        self.shot_action = False
        self.all_sprites_list.update()

        if action % 3 == 0:
            self.ship.yspd = 0
        if action % 3 == 1:
            self.ship.yspd = -self.ship.speed
        if action % 3 == 2:
            self.ship.yspd = self.ship.speed
        if action >= 3 and action < 6:
            self.ship.rotate_ship(self.ship.rotspd)
        if action >= 6 and action < 9:
            self.ship.rotate_ship(-self.ship.rotspd)
        if action == 9:
            if self.ammo > 0:
                bullet = Bullet(self.projectile_sprites, (20, 20), self)
                self.bullets.add(bullet)
                self.all_sprites_list.add(bullet)
                self.all_sprites_list.remove(self.ship)
                self.all_sprites_list.add(self.ship)
                self.ammo -= 1
                self.shot = True
            else:
                self.shot_action = True

        if self.frame % self.WALL_INTERVAL == 0 and self.WALL_SPAWN:
            y = 0
            if self.ENEMYSPAWN == "random":
                y = self.np_random.integers(0, self.screen_height - 100)
            else:
                y = self.genY
            wall = Wall(self.wall_sprites, self.screen_width, y, 20, 100, self)
            self.genY = (
                self.genY + (50 if self.ENEMYSPAWN == "down" else -50)
            ) % self.screen_height
            self.walls.add(wall)
            self.all_sprites_list.add(wall)
        if self.frame % self.SQUID_INTERVAL == 0:
            y = 0
            if self.ENEMYSPAWN == "random":
                y = self.np_random.integers(0, self.screen_height - 100)
            else:
                y = self.genY
            squid_color = self.np_random.choice(["ORANGE", "PINK"])
            squid = Squid(
                self.squid_sprites[squid_color], self.screen_width, y, 115, 40, self
            )
            self.genY = (
                self.genY + (50 if self.ENEMYSPAWN == "down" else -50)
            ) % self.screen_height
            self.squids.add(squid)
            self.all_sprites_list.add(squid)
        if self.frame % self.AMMO_INTERVAL == 0 and self.ammo < self.MAX_AMMO:
            self.ammo += 1
        self.check_collisions()
        self.clear_sprites()
        self.time_text.update("TIME: " + str(round(self.TIME_LIMIT - self.time, 1)))
        self.ammo_text.update("BULLETS: " + str(self.ammo))
        self.score_text.update("SCORE: " + str(self.score))
        if self.TIME_LIMIT - self.time <= 0:
            self.done = True

        self.city_x_position -= self.city_scroll_speed
        if self.city_x_position <= -self.screen_width:
            self.city_x_position = 0

    def clear_sprites(self):
        for i in self.walls_to_remove:
            self.walls.remove(i)
            self.all_sprites_list.remove(i)
        self.walls_to_remove = set()
        for i in self.squids_to_remove:
            self.squids.remove(i)
            self.all_sprites_list.remove(i)
        self.squids_to_remove = set()
        for i in self.bullets_to_remove:
            self.bullets.remove(i)
            self.all_sprites_list.remove(i)
        self.bullets_to_remove = set()

    def check_collisions(self):
        for i in self.walls.union(self.squids):
            if pygame.Rect(
                self.ship.rect.x,
                self.ship.rect.y,
                self.ship.image.get_width(),
                self.ship.image.get_height(),
            ).colliderect(i.rect):
                self.done = True
                self.dead = True
                return

            for j in self.bullets:
                if i.rect.colliderect(j.rect):
                    if i in self.squids:
                        i.explode(self.explosion_sprites, (100, 100))
                        self.killed = True
                        self.score += 1
                    self.bullets_to_remove.add(j)
                    break

    def get_ray_distance(self, clip):
        buffer = 150
        dist = 1 - (clip - buffer) / (self.screen_width - buffer)
        if dist > 1:
            return -1
        return dist

    def get_observation(self):
        """
        Return the current observation from the game state.
        """
        # use the game state to output an array in the format of your observation space
        wall_casts = []
        squid_casts = []
        y = 25
        # horizontal rays that return the x coordinate of the leftmost intersecting wall and squid
        while y < self.screen_height:
            clipped = False
            for i in self.walls:
                clip = i.rect.clipline(0, y, self.screen_width, y)
                if clip != ():
                    pygame.draw.circle(self.screen, RED, clip[0], 3)
                    wall_casts.append(self.get_ray_distance(clip[0][0]))
                    clipped = True
                    break
            if not clipped:
                wall_casts.append(-1)
            clipped = False
            for i in self.squids:
                clip = i.rect.clipline(0, y, self.screen_width, y)
                if clip != ():
                    pygame.draw.circle(self.screen, RED, clip[0], 3)
                    squid_casts.append(self.get_ray_distance(clip[0][0]))
                    clipped = True
                    break
            if not clipped:
                squid_casts.append(-1)
            y += 50

        ship_position = ((self.ship.y + SHIP_SIZE / 2) / self.screen_height - 0.5) * 2
        ship_position /= 1 - SHIP_SIZE / self.screen_height

        if self.ship.angle > 180:
            angle_feature = (self.ship.angle - 360) / SHIP_MAX_ANGLE
        else:
            angle_feature = self.ship.angle / SHIP_MAX_ANGLE
        arr = (
            [ship_position, angle_feature, self.ammo / self.MAX_AMMO]
            + wall_casts
            + squid_casts
        )
        return np.array(arr, dtype=np.float32)

    def get_nearest_hazard(self):
        mini = 9999
        minx = 9999
        for i in self.squids.union(self.walls):
            proxx = (
                i.rect.x
                + i.image.get_width() / 2
                - self.ship.rect.x
                - self.ship.image.get_width() / 2
            )
            if 0 <= proxx < minx:
                prox = abs(
                    i.rect.y
                    + i.image.get_height() / 2
                    - self.ship.rect.y
                    - self.ship.image.get_height() / 2
                )
                mini = min(mini, prox)
                minx = (
                    i.rect.x
                    + i.image.get_width() / 2
                    - self.ship.rect.x
                    - self.ship.image.get_width() / 2
                )
        return mini

    def _classic_reward(self):
        reward = 0
        if self.killed:
            reward += 0.5
        if self.dead:
            reward -= 5
        return reward + 0.01

    def _highest_kill_count_reward(self):
        reward = 0
        if self.killed:
            reward += 1
        if self.shot and not self.killed:  # Missing a shot
            reward -= 0.01
        return reward + 0.01

    def _precision_shooting_reward(self):
        reward = 0
        if self.killed:
            self.combo_streak += 1
            reward += 0.5 * self.combo_streak
        elif self.shot and not self.killed:  # Missing a shot
            reward -= 0.05
            self.combo_streak = 0  # Reset combo
        return reward + 0.01

    def _ammo_efficiency_reward(self):
        reward = 0
        if self.killed:
            reward += 1
        if self.shot:  # Penalty for each shot fired
            reward -= 0.1
        return reward + 0.01

    def _time_survival_reward(self):
        reward = 0
        if self.dead:
            reward -= 5
        return reward + 0.01



from gymnasium import register

__version__ = "0.1.0"

register(
    id="CoopPuzzle-v0",
    entry_point="sai_pygame.coop_puzzle:CoopPuzzleEnv",
)

register(
    id="Pong-v0",
    entry_point="sai_pygame.pong:PongEnv",
)

register(
    id="SpaceEvaders-v0",
    entry_point="sai_pygame.space_evaders:SpaceEvadersEnv",
)

register(
    id="SquidHunt-v0",
    entry_point="sai_pygame.squid_hunt:SquidHuntEnv",
)

## Import Future Environments if Available
try:
    import sai_pygame._future
except:
    pass



import math

from sai_pygame.utils.animation import AnimationBase

from ..constants.config import SHIP_SIZE, SCREEN_HEIGHT, SCREEN_WIDTH


class Bullet(AnimationBase):
    def __init__(self, projectile_sprites, size, game):
        self.game = game
        x_offset = SHIP_SIZE * 0.5
        y_offset = SHIP_SIZE * 0.3
        self.pos_offset = math.sqrt(x_offset**2 + y_offset**2)
        self.pos_angle_offset = math.atan(y_offset / x_offset)

        x = game.ship.center_pos[0] + self.pos_offset * math.cos(
            math.radians(game.ship.angle) - self.pos_angle_offset
        )
        y = game.ship.center_pos[1] - self.pos_offset * math.sin(
            math.radians(game.ship.angle) - self.pos_angle_offset
        )

        super().__init__(projectile_sprites, (x, y), size, 3, [5, 5, 5])
        self.spd = 8
        self.xspd = 3
        self.yspd = 0
        self.rotation = self.game.ship.angle

    def update(self):
        self.xspd = math.cos(math.radians(self.rotation)) * self.spd
        self.yspd = math.sin(math.radians(self.rotation)) * -self.spd
        self.rect.x = self.rect.x + self.xspd
        self.rect.y = self.rect.y + self.yspd
        if (
            self.rect.x >= SCREEN_WIDTH
            or self.rect.x <= 0
            or self.rect.y >= SCREEN_HEIGHT
            or self.rect.y <= 0
        ):
            self.game.bullets_to_remove.add(self)

        return super().update(stay_on_last_frame=True)



from sai_pygame.utils.colors import WHITE
from sai_pygame.utils.fonts import load_font, fonts


class Text:
    def __init__(self, x, y, string, size, game):
        super().__init__()
        self.x = x
        self.y = y
        self.size = size
        self.game = game
        self.screen_width = game.screen_width
        self.screen_height = game.screen_height
        self.update(string)

    def update(self, string):
        self.image = load_font(fonts["thick"], self.size).render(string, 1, WHITE)



import pygame
import os

from ..assets import GAME_ASSETS_BASE
from ..constants import SCREEN_HEIGHT, SCREEN_WIDTH, SHIP_SIZE, SHIP_MAX_ANGLE
from .ship_exhaust import ShipExhaust


class Ship(pygame.sprite.Sprite):
    def __init__(self, exhaust_sprites):
        super().__init__()
        self.x = 0
        self.y = 0
        self.center_pos = (0, 0)
        self.xspd = 0
        self.yspd = 0
        self.speed = 6
        self.rotspd = 4
        self.angle = 0

        self.img = pygame.image.load(os.path.join(GAME_ASSETS_BASE, "ship.png"))
        self.img = pygame.transform.scale(self.img, (SHIP_SIZE, SHIP_SIZE))
        self.image = pygame.transform.rotate(self.img, self.angle)

        self.rect = pygame.Rect(100, 100, SHIP_SIZE, SHIP_SIZE)
        self.exhaust = ShipExhaust(self, exhaust_sprites, (0, 0))

    def reset(self):
        self.x = SCREEN_WIDTH / 4 - SHIP_SIZE / 2
        self.y = SCREEN_HEIGHT / 2 - SHIP_SIZE / 2
        self.angle = 0
        self.yspd = 0
        self.xspd = 0

    def rot_center(self, image, angle):
        rotated_image = pygame.transform.rotate(image, angle)
        new_rect = rotated_image.get_rect(center=image.get_rect().center)
        self.center_pos = (new_rect.center[0] + self.x, new_rect.center[1] + self.y)
        self.rect.x = new_rect.x + self.x
        self.rect.y = new_rect.y + self.y
        self.image = rotated_image

    def rotate_ship(self, angle):
        prev_angle = self.angle
        self.angle += angle
        if self.angle >= SHIP_MAX_ANGLE and self.angle <= 360 - SHIP_MAX_ANGLE:
            self.angle = prev_angle

    def update(self):
        self.angle %= 360
        self.x = max(0, min(self.x + self.xspd, SCREEN_WIDTH - SHIP_SIZE))
        self.y = max(0, min(self.y + self.yspd, SCREEN_HEIGHT - SHIP_SIZE))
        self.rot_center(self.img, self.angle)
        self.exhaust.update(self)



import pygame
from ..constants import SCREEN_WIDTH


class Wall(pygame.sprite.Sprite):
    def __init__(self, wall_sprites, x, y, w, h, game):
        super().__init__()
        self.wall_sprites = wall_sprites
        self.x = x
        self.y = y
        self.w = w
        self.h = h
        self.xspd = -3
        self.yspd = 0
        self.game = game
        self.image = pygame.Surface([w, h])
        self.image.fill((0, 0, 255))

        self.image = self.wall_sprites[0]
        self.image = pygame.transform.scale(self.image, (self.w, self.h))
        self.rect = self.image.get_rect(center=(self.x, self.y))

    def update(self):
        self.rect.x = self.rect.x + self.xspd
        self.rect.y = self.rect.y + self.yspd

        if self.rect.x > SCREEN_WIDTH * 0.66:
            self.image = self.wall_sprites[0]
        elif self.rect.x > SCREEN_WIDTH * 0.33:
            self.image = self.wall_sprites[1]
        else:
            self.image = self.wall_sprites[2]

        self.image = pygame.transform.scale(self.image, (self.w, self.h))

        if self.rect.x < 0 - self.w:
            self.game.walls_to_remove.add(self)



FRAMERATE = 60
SCREEN_HEIGHT = 480
SCREEN_WIDTH = 1200

BACKGROUND_SPEED = 0.5
BACKGROUND_OPACITY = 125  # 0-255
BACKGROUND_WIDTH = 1700
BACKGROUND_HEIGHT = 600

TIME_LIMIT = 30

SHIP_SIZE = 75
SHIP_MAX_ANGLE = 45



import pygame
import os

fonts = {"thick": "retro-pixel-thick.ttf", "arcade": "retro-pixel-arcade.ttf"}


def load_font(font_name, size):
    font_path = os.path.join(os.path.dirname(__file__), font_name)
    return pygame.font.Font(font_path, size)



import pygame
import os


def get_tile(sheet, tile_width, tile_height, tile_row, tile_column):
    x = tile_column * tile_width
    y = tile_row * tile_height
    return sheet.subsurface((x, y, tile_width, tile_height))


def load_spritesheet(filename, num_tiles, mapping):
    image_path = os.path.join(os.path.dirname(__file__), filename)
    sheet = pygame.image.load(image_path)
    sheet = pygame.transform.scale(sheet, (num_tiles[0] * 40, num_tiles[1] * 40))
    sheet_width, sheet_height = sheet.get_size()
    tile_width = sheet_width // num_tiles[0]
    tile_height = sheet_height // num_tiles[1]
    sprite_sheet = {}
    for data in mapping:
        sprite_sheet[data[2]] = get_tile(
            sheet, tile_width, tile_height, data[1], data[0]
        )
    return sprite_sheet


def load_sprites_from_separate_files(file_prepend, num_sprites):
    sprite_sheet = {}
    for i in range(num_sprites):
        image_path = os.path.join(
            os.path.dirname(__file__), "{}-{}.png".format(file_prepend, i + 1)
        )
        sprite_sheet[i] = pygame.image.load(image_path)
    return sprite_sheet
